/*
                                  MatchWild

                                by Joe Strout
MatchWild is not a class; instead, it is a simple module containing four
functions:

   * int MatchWild(s,p): return 1 if s matches p, where p may contain
     wildcards

In these functions, all parameters are character pointers (const where
appropriate) except n, which is the length of a buffer to hold the result of
a replace operation. (It's OK for this buffer to be the same as one of the
source strings.)

The only currently supported wildcard is the asterisk, which matches zero or
more characters.

  ------------------------------------------------------------------------

Examples

MatchWild( "this is a test", "this*") returns 1
MatchWild( "is this a test", "this*") returns 0
MatchWild( "is this a test", "*this*") returns 1
  ------------------------------------------------------------------------

Here's the code...

   * MatchWild.h (header file)
   * MatchWild.cpp (source code)

Share and enjoy!
http://www-ncmir.ucsd.edu/~jstrout/classlib/MatchWild.html

Last Updated: 9/23/96 Joe Strout ( jstrout@ucsd.edu )

*/

/*****************************************************************************/
/*  Includes                                                                 */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*****************************************************************************/
/*  Macros, typedefs                                                         */
/*****************************************************************************/


/*****************************************************************************/
/*  Global Vars                                                              */
/*****************************************************************************/

const int MAXPIECES = 257;          /* maximum non-wild parts we can handle  */
const int MAXWILDS = MAXPIECES+2;   /* maximum wildcards  */
static char *wilds[MAXPIECES];      /* parts of given string which
                                       fit in wildcards */
static int numWilds=0;              /* how many wild parts we have */


/*****************************************************************************/
/*  Prototypes                                                               */
/*****************************************************************************/

int MatchWild( char *pStr,char *pPat );
void ClearWilds(void);
void AddWild(char *pStart,int pNumChars );
int CountNonwild( char *pPtr );
int FindSubstr(char *pStr,char *pSubstr );


/*****************************************************************************/
/*  Functions                                                                */
/*****************************************************************************/

void ClearWilds(void)
{
int i;

    for (i=0; i<numWilds; i++)
        if (wilds[i]) free (wilds[i]);
    numWilds = 0;
}


int FindSubstr(char *pStr, char *pSubstr )
{
    char *p = pStr;
    char *maxp;
    int len = strlen(pSubstr);

    /* return the offset of the first occurance of pSubstr in pStr,
       or -1 if not found */
    if (!*pSubstr) return 0;

    maxp = pStr + strlen(pStr) - len;
    for (; p <= maxp; p++) {
        if (!strncmp( p, pSubstr, len ) ) return p-pStr;    /* found it! */
    }
    return -1;
}




int CountNonwild(char *pPtr )
{
    /* return how many chars at the beginning of pPtr are not wildcards */
    const char *p=pPtr;
    for (; *p && *p != '*'; p++) {}
    return p-pPtr;
}



void AddWild( char *pStart,int pNumChars )
{
    if (numWilds==MAXWILDS || !pNumChars) return;
    wilds[numWilds] = (char *) malloc(pNumChars+1);
    strncpy( wilds[numWilds], pStart, pNumChars );
    wilds[numWilds][pNumChars] = '\0';
    numWilds++;
}



int MatchWild(char *pStr, char *pPat )
{
    /* first, break the pattern string (pPat) into non-wildcard pieces */
    char *pieces[MAXPIECES];        /* array of tame pieces */
    int pieceLen[MAXPIECES];        /* the length of each pieces */
    int patLen;                     /* total length of the pattern string */
    int numPieces = 0;              /* number of pieces found */
    char *p = pPat;                 /* temporary pointer into pPat */
    char *s = pStr;                 /* temporary pointer into pStr */
    int offset,i;                   /* offset of each matching substring */

    if ((pPat[0] == '*') && (pPat[1] == '\0'))
         return true;               /* match everything */
    
    patLen = strlen(pPat);
    ClearWilds();                   /* clear out wildcard match holders */

    while (p < pPat+patLen && numPieces < MAXPIECES) {
        /* get length of nonwild piece */
        pieceLen[numPieces] = CountNonwild(p);
        if (pieceLen[numPieces]) {          /* if we have such a pieces... */
            pieces[numPieces] = (char *) malloc(pieceLen[numPieces]+1);
            /* to copy it */
            strncpy( pieces[numPieces], p, pieceLen[numPieces] );
            /* end in null */
            pieces[numPieces][pieceLen[numPieces]] = '\0';
            p += pieceLen[numPieces] + 1;       /* advance the pointer */
            numPieces++;                        /* increase piece count */
        }
        else p++;   /* if piece size was zero, just advance pointer 1 char */
    }
    if (numPieces == MAXPIECES) return 0; /* whoops -- exceeded capacity! */

    /* now find the first occurrance in pStr of each piece */


    for (i=0; i<numPieces; i++) {   /* for each piece... */
        offset = FindSubstr( s, pieces[i] );    /* look for a match in s */
        if (offset < 0) return 0;        /* piece not found -- no match! */
        if (!i && pPat[0] != '*' && offset > 0) {
            /* unaccounted-for chars at beginning of string -- no match */
            return 0;
        }
        /* copy the stuff that matched the wildcard... */
        AddWild( s, offset );
        s += offset + pieceLen[i]; /* advance pointer to look for next piece */
    }

    /* check end of string */
    if (pPat[patLen-1] == '*') {
        AddWild( s, pStr+strlen(pStr)-s );  /* copy one last wild */
    }
    else {
        /* if no wildcard at end of string, then we'd
           better have munched it all... */
        if (s < pStr + strlen(pStr)) return 0;/* nope, unaccounted-for chars */
    }

    return 1;
}
